/*
  ThingsBoardHttp.h - Library API for sending data to the ThingsBoard
  Based on PubSub MQTT library.
  Created by Olender M. Oct 2018.
  Released into the public domain.
*/
#ifndef ThingsBoard_Http_h
#define ThingsBoard_Http_h

// Library includes.
#if !defined(ESP8266) && !defined(ESP32)
#include <ArduinoHttpClient.h>
#else
#include <HTTPClient.h>
#endif

// Local includes.
#include "ThingsBoardDefaultLogger.h"
#include "Telemetry.h"
#include "Constants.h"

/// ---------------------------------
/// Constant strings in flash memory.
/// ---------------------------------
// HTTP topics.
constexpr char HTTP_TELEMETRY_TOPIC[] PROGMEM = "/api/v1/%s/telemetry";
constexpr char HTTP_ATTRIBUTES_TOPIC[] PROGMEM = "/api/v1/%s/attributes";
constexpr char HTTP_POST_PATH[] PROGMEM = "application/json";

// Log messages.
constexpr char POST[] PROGMEM = "POST";
constexpr char GET[] PROGMEM = "GET";
constexpr char HTTP_FAILED[] PROGMEM = "(%s) failed HTTP response (%d)";

/// @brief Wrapper around the ArduinoHttpClient or HTTPClient to allow connecting and sending / retrieving data from ThingsBoard over the HTTP orHTTPS protocol
/// @tparam Logger Logging class that should be used to print messages generated by ThingsBoard, default = ThingsBoardDefaultLogger
/// @tparam MaxFieldsAmt Maximum amount of key value pair that we will be able to sent to ThingsBoard in one call, default = 8
template<size_t MaxFieldsAmt = Default_Fields_Amt,
         typename Logger = ThingsBoardDefaultLogger>
class ThingsBoardHttpSized {
  public:
    /// @brief Initalizes the underlying client with the needed information
    /// so it can initally connect to the given host over the given port
    /// @param client Client that should be used to establish the connection
    /// @param access_token Token used to verify the devices identity with the ThingsBoard server
    /// @param host Host server we want to establish a connection to (example: "demo.thingsboard.io")
    /// @param port Port we want to establish a connection over (80 for HTTP, 443 for HTTPS)
#if !defined(ESP8266) && !defined(ESP32)
    inline ThingsBoardHttpSized(Client &client, const char *access_token,
                                const char *host, const uint16_t& port = 80U)
      : m_client(client, host, port)
#else
    inline ThingsBoardHttpSized(WiFiClient &client, const char *access_token,
                                const char *host, const uint16_t& port = 80U)
      : m_client()
      , m_wifiClient(client)
      , m_secure(port == 433U)
#endif
      , m_host(host)
      , m_port(port)
      , m_token(access_token)
    {
#if defined(ESP8266) || defined(ESP32)
      m_client.addHeader("Content-Type", HTTP_POST_PATH);
      m_client.begin(client, m_host, m_port, "/", m_secure);
#else
      m_client.connect(m_host, m_port)
#endif
    }

    /// @brief Destructor
    inline ~ThingsBoardHttpSized() {
      // Nothing to do.
    }

    //----------------------------------------------------------------------------
    // Telemetry API

    /// @brief Attempts to send integer telemetry data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetryInt(const char *key, int value) {
      return sendKeyValue(key, value);
    }

    /// @brief Attempts to send boolean telemetry data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetryBool(const char *key, bool value) {
      return sendKeyValue(key, value);
    }

    /// @brief Attempts to send float telemetry data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetryFloat(const char *key, float value) {
      return sendKeyValue(key, value);
    }

    /// @brief Attempts to send string telemetry data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetryString(const char *key, const char *value) {
      return sendKeyValue(key, value);
    }

    /// @brief Attempts to send aggregated telemetry data
    /// @param data Array containing all the data we want to send
    /// @param data_count Amount of data entries in the array that we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetry(const Telemetry *data, size_t data_count) {
      return sendDataArray(data, data_count);
    }

    /// @brief Attempts to send custom json telemetry string
    /// @param json String containing our json key value pairs we want to attempt to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetryJson(const char *json) {
      if (json == nullptr || m_token == nullptr) {
        return false;
      }

      char path[detectSize(HTTP_TELEMETRY_TOPIC, m_token)];
      snprintf_P(path, sizeof(path), HTTP_TELEMETRY_TOPIC, m_token);
      return postMessage(path, json);
    }

    /// @brief Attempts to send custom telemetry JsonObject
    /// @param jsonObject JsonObject containing our json key value pairs we want to send
    /// @param jsonSize Size of the data inside the JsonObject
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetryJson(const JsonObject& jsonObject, const uint32_t& jsonSize) {
      const uint32_t jsonObjectSize = jsonObject.size();
      if (MaxFieldsAmt < jsonObjectSize) {
        char message[detectSize(TOO_MANY_JSON_FIELDS, jsonObjectSize, MaxFieldsAmt)];
        snprintf_P(message, sizeof(message), TOO_MANY_JSON_FIELDS, jsonObjectSize, MaxFieldsAmt);
        Logger::log(message);
        return false;
      }
      char json[jsonSize];
      // Serialize json does not include size of the string null terminator
      if (serializeJson(jsonObject, json, jsonSize) < jsonSize - 1) {
        Logger::log(UNABLE_TO_SERIALIZE_JSON);
        return false;
      }
      return sendTelemetryJson(json);
    }

    /// @brief Attempts to send custom telemetry JsonVariant
    /// @param jsonVariant JsonVariant containing our json key value pairs we want to send
    /// @param jsonSize Size of the data inside the JsonVariant
    /// @return Wheter sending the data was successful or not
    inline const bool sendTelemetryJson(const JsonVariant& jsonVariant, const uint32_t& jsonSize) {
      const uint32_t jsonVariantSize = jsonVariant.size();
      if (MaxFieldsAmt < jsonVariantSize) {
        char message[detectSize(TOO_MANY_JSON_FIELDS, jsonVariantSize, MaxFieldsAmt)];
        snprintf_P(message, sizeof(message), TOO_MANY_JSON_FIELDS, jsonVariantSize, MaxFieldsAmt);
        Logger::log(message);
        return false;
      }
      char json[jsonSize];
      // Serialize json does not include size of the string null terminator
      if (serializeJson(jsonVariant, json, jsonSize) < jsonSize - 1) {
        Logger::log(UNABLE_TO_SERIALIZE_JSON);
        return false;
      }
      return sendTelemetryJson(json);
    }

    //----------------------------------------------------------------------------
    // Attribute API

    /// @brief Attempts to send integer attribute data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributeInt(const char *attrName, int value) {
      return sendKeyValue(attrName, value, false);
    }

    /// @brief Attempts to send boolean attribute data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributeBool(const char *attrName, bool value) {
      return sendKeyValue(attrName, value, false);
    }

    /// @brief Attempts to send float attribute data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributeFloat(const char *attrName, float value) {
      return sendKeyValue(attrName, value, false);
    }

    /// @brief Attempts to send string attribute data with the given key and value
    /// @param key Key of the key value pair we want to send
    /// @param value Value of the key value pair we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributeString(const char *attrName, const char *value) {
      return sendKeyValue(attrName, value, false);
    }

    /// @brief Attempts to send aggregated attribute data
    /// @param data Array containing all the data we want to send
    /// @param data_count Amount of data entries in the array that we want to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributes(const Attribute *data, size_t data_count) {
      return sendDataArray(data, data_count, false);
    }

    /// @brief Attempts to send custom json attribute string
    /// @param json String containing our json key value pairs we want to attempt to send
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributeJSON(const char *json) {
      if (json == nullptr || m_token == nullptr) {
        return false;
      }

      char path[detectSize(HTTP_ATTRIBUTES_TOPIC, m_token)];
      snprintf_P(path, sizeof(path), HTTP_ATTRIBUTES_TOPIC, m_token);
      return postMessage(path, json);
    }

    /// @brief Attempts to send custom attribute JsonObject
    /// @param jsonObject JsonObject containing our json key value pairs we want to send
    /// @param jsonSize Size of the data inside the JsonObject
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributeJSON(const JsonObject& jsonObject, const uint32_t& jsonSize) {
      const uint32_t jsonObjectSize = jsonObject.size();
      if (MaxFieldsAmt < jsonObjectSize) {
        char message[detectSize(TOO_MANY_JSON_FIELDS, jsonObjectSize, MaxFieldsAmt)];
        snprintf_P(message, sizeof(message), TOO_MANY_JSON_FIELDS, jsonObjectSize, MaxFieldsAmt);
        Logger::log(message);
        return false;
      }
      char json[jsonSize];
      // Serialize json does not include size of the string null terminator
      if (serializeJson(jsonObject, json, jsonSize) < jsonSize - 1) {
        Logger::log(UNABLE_TO_SERIALIZE_JSON);
        return false;
      }
      return sendAttributeJSON(json);
    }

    /// @brief Attempts to send custom attribute JsonVariant
    /// @param jsonVariant JsonVariant containing our json key value pairs we want to send
    /// @param jsonSize Size of the data inside the JsonVariant
    /// @return Wheter sending the data was successful or not
    inline const bool sendAttributeJSON(const JsonVariant& jsonVariant, const uint32_t& jsonSize) {
      const uint32_t jsonVariantSize = jsonVariant.size();
      if (MaxFieldsAmt < jsonVariantSize) {
        char message[detectSize(TOO_MANY_JSON_FIELDS, jsonVariantSize, MaxFieldsAmt)];
        snprintf_P(message, sizeof(message), TOO_MANY_JSON_FIELDS, jsonVariantSize, MaxFieldsAmt);
        Logger::log(message);
        return false;
      }
      char json[jsonSize];
      // Serialize json does not include size of the string null terminator
      if (serializeJson(jsonVariant, json, jsonSize) < jsonSize - 1) {
        Logger::log(UNABLE_TO_SERIALIZE_JSON);
        return false;
      }
      return sendAttributeJSON(json);
    }

  private:
    /// @brief Attempts to establishes a connection over HTTP or HTTPS
    /// @param path API path we want to connect over (example: /api/v1/$TOKEN/attributes)
    /// @return Wheter connecting to the host with the given API path was successful or not
    inline const bool connect(const char* path) {
      if (path == nullptr) {
        return false;
      }
#if !defined(ESP8266) || defined(ESP32)
      m_client.setURL(path);
#endif

      if (m_client.connected()) {
        return true;
      }

#if !defined(ESP8266) && !defined(ESP32)
      if (!m_client.connect(m_host, m_port)) {
#else
      if (!m_client.begin(m_wifiClient, m_host, m_port, path, m_secure)) {
#endif
        Logger::log(CONNECT_FAILED);
        return false;
      }
      return true;
    }

    /// @brief Clears any remaining memory of the previous conenction,
    /// but keeps the TCP connection open so it can be reused,
    /// without needing to re-establish the TCP handshake again
    inline void clearConnection() {
#if !defined(ESP8266) && !defined(ESP32)
      m_client.stop();
#else
      m_client.end();
#endif
    }

    /// @brief Attempts to send a POST request over HTTP or HTTPS
    /// @param path API path we want to send data to (example: /api/v1/$TOKEN/attributes)
    /// @param json String containing our json key value pairs we want to attempt to send
    /// @return Wheter sending the POST request was successful or not
    inline const bool postMessage(const char* path, const char* json) {
      if (!connect(path)) {
        return false;
      }

      bool result = true;

#if !defined(ESP8266) && !defined(ESP32)
      const bool success = m_client.post(path, HTTP_POST_PATH, json);
      const int status = m_client.responseStatusCode();
      if (!success || status != HTTP_SUCCESS) {
#else
      const int status = m_client.POST(json);
      if (status != HTTP_CODE_OK) {
#endif
        char message[detectSize(HTTP_FAILED, POST, status)];
        snprintf_P(message, sizeof(message), HTTP_FAILED, POST, status);
        Logger::log(message);
        result = false;
      }

      clearConnection();
      return result;
    }

    /// @brief Attempts to send a GET request over HTTP or HTTPS
    /// @param path API path we want to get data from (example: /api/v1/$TOKEN/rpc)
    /// @param response JsonObject the GET response will be copied into,
    /// will not be changed if the GET request wasn't successful
    /// @return Wheter sending the GET request was successful or not
    inline const bool getMessage(const char* path, JsonObject& response) {
      if (!connect(path)) {
        return false;
      }

      bool result = true;
      DeserializationError error(DeserializationError::Code::Ok);
      const char* payload = nullptr;
      static StaticJsonDocument<JSON_OBJECT_SIZE(MaxFieldsAmt)> jsonBuffer;
      jsonBuffer.clear();

#if !defined(ESP8266) && !defined(ESP32)
      const bool success = m_client.get(path);
      const int status = m_client.responseStatusCode();
      if (!success || status != HTTP_SUCCESS) {
#else
      const int status = m_client.GET();
      if (status != HTTP_CODE_OK) {
#endif
        char message[detectSize(HTTP_FAILED, GET, status)];
        snprintf_P(message, sizeof(message), HTTP_FAILED, GET, status);
        Logger::log(message);
        result = false;
        goto cleanup;
      }

#if !defined(ESP8266) && !defined(ESP32)
      payload = m_client.responseBody().c_str();
#else
      payload = m_client.getString().c_str();
#endif
      error = deserializeJson(jsonBuffer, payload);

      if (error) {
        char message[detectSize(UNABLE_TO_DE_SERIALIZE_JSON, error.c_str())];
        snprintf_P(message, sizeof(message), UNABLE_TO_DE_SERIALIZE_JSON, error.c_str());
        Logger::log(message);
        result = false;
        goto cleanup;
      }

      response = jsonBuffer.template as<JsonObject>();

      cleanup:
      clearConnection();
      return result;
    }

    /// @brief Detects the size of the given format string combined with the given arguments
    /// @param msg Format of the message we want to print, same syntax as printf
    /// @param ... Additional parameters passed that will be inserted in the given format message
    /// @return Returns the length in characters needed for a given format string with the given arguments to be displayed completly
    inline const uint8_t detectSize(const char* msg, ...) const {
      va_list args;
      va_start(args, msg);
      // Result is what would have been written if the passed buffer would have been large enough not counting null character,
      // or if an error occured while creating the string a negative number is returned instead. TO ensure this will not crash the system
      // when creating an array with negative size we assert beforehand with a clear error message.
      const int32_t result = JSON_STRING_SIZE(vsnprintf_P(nullptr, 0U, msg, args));
      assert(result >= 0);
      va_end(args);
      return result;
    }

    /// @brief Attempts to send aggregated attribute or telemetry data
    /// @param data Array containing all the data we want to send
    /// @param data_count Amount of data entries in the array that we want to send
    /// @param telemetry Wheter the aggregated data is telemetry (true) or attribut (false)
    /// @return Wheter sending the data was successful or not
    inline const bool sendDataArray(const Telemetry *data, size_t data_count, bool telemetry = true) {
      StaticJsonDocument<JSON_OBJECT_SIZE(MaxFieldsAmt)> jsonBuffer;
      JsonVariant object = jsonBuffer.template to<JsonVariant>();

      for (size_t i = 0; i < data_count; ++i) {
        if (!data[i].SerializeKeyValue(object)) {
          Logger::log(UNABLE_TO_SERIALIZE);
          return false;
        }
      }

      return telemetry ? sendTelemetryJson(object, JSON_STRING_SIZE(measureJson(object))) : sendAttributeJSON(object, JSON_STRING_SIZE(measureJson(object)));
    }

    /// @brief Sends single key-value attribute or telemetry data in a generic way
    /// @tparam T Type of the passed value
    /// @param key Key of the key value pair we want to send
    /// @param val Value of the key value pair we want to send
    /// @param telemetry Wheter the aggregated data is telemetry (true) or attribut (false)
    /// @return Wheter sending the data was successful or not
    template<typename T>
    inline const bool sendKeyValue(const char *key, T value, bool telemetry = true) {
      const Telemetry t(key, value);
      if (t.IsEmpty()) {
        // Message is ignored and not sent at all.
        return false;
      }

      StaticJsonDocument<JSON_OBJECT_SIZE(1)> jsonBuffer;
      JsonVariant object = jsonBuffer.template to<JsonVariant>();
      if (!t.SerializeKeyValue(object)) {
        Logger::log(UNABLE_TO_SERIALIZE);
        return false;
      }

      return telemetry ? sendTelemetryJson(object, JSON_STRING_SIZE(measureJson(object))) : sendAttributeJSON(object, JSON_STRING_SIZE(measureJson(object)));
    }

#if !defined(ESP8266) && !defined(ESP32)
    HttpClient m_client;
#else
    HTTPClient m_client;
    WiFiClient& m_wifiClient;
    const bool m_secure;
#endif
    const char *m_host;
    const uint16_t m_port;
    const char *m_token;
};

using ThingsBoardHttp = ThingsBoardHttpSized<>;

#endif // ThingsBoard_Http_h
